# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231122

import functools
import inspect
from ast import literal_eval
from threading import Thread
from typing import Callable, Optional

from neetbox._protocol import *
from neetbox.logging import logger
from neetbox.utils import Registry
from neetbox.utils.mvc import Singleton

from .._client import connection


class Action(Callable):
    def __init__(
        self,
        function: Callable,
        name: str = None,
        description: str = None,
        blocking: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.function = function
        self.name = name if name else function.__name__
        self.description = description
        self.argspec = inspect.getfullargspec(self.function)
        self.blocking = blocking

    def get_props_dict(self):
        # _arg_dict = {
        #     _arg_name: self.argspec.annotations.get(_arg_name, None)
        #     for _arg_name in self.argspec.args
        # }
        _arg_anno_dict = self.function.__annotations__
        _args = self.argspec.args
        _arg_dict = {}
        for _arg_name in _args:
            _arg_type = _arg_anno_dict.get(_arg_name, any)
            _arg_dict[_arg_name] = _arg_type if isinstance(_arg_type, str) else _arg_type.__name__
        return {
            "description": self.description,
            "args": _arg_dict,
            "blocking": self.blocking,
        }

    def __call__(self, **argv):
        self.function(argv)  # ignore blocking

    def eval_call(self, params: dict):
        eval_params = {}
        for k, v in params.items():
            if v:
                eval_params[k] = literal_eval(v)
        return self.function(**eval_params)


class ActionManager(metaclass=Singleton):
    ACTION_POOL: Registry = Registry("__NEET_ACTIONS")
    _is_initialized = False

    def get_action_dict(self):
        return {name: self.ACTION_POOL[name].get_props_dict() for name in self.ACTION_POOL.keys()}

    def eval_call(self, name: str, params: dict, callback: Optional[Callable] = None):
        if name not in self.ACTION_POOL:
            logger.err(f"Could not find action with name {name}, action stopped.")
            return False
        target_action: Action = self.ACTION_POOL[name]
        logger.log(
            f"Agent runs function '{target_action.name}', blocking = {target_action.blocking}"
        )

        def run_and_callback():
            try:
                returned_data = target_action.eval_call(params)
            except Exception as e:
                returned_data = e
                logger.warn(f"action {target_action} failed with exception {e}")
            if callback:
                callback(returned_data)

        if not target_action.blocking:  # non-blocking run in thread
            Thread(
                target=run_and_callback,
                daemon=True,
            ).start()
            return
        else:  # blocking run
            run_and_callback()
            return

    def _initialize(self):
        if not self._is_initialized:

            @connection.ws_subscribe(event_type_name=EVENT_TYPE_NAME_ACTION)
            def _listen_to_actions(message: EventMsg):
                actionManager.eval_call(
                    name=message.payload[NAME_KEY],
                    params=message.payload[ARGS_KEY],
                    callback=lambda x: connection.ws_send(
                        event_type=EVENT_TYPE_NAME_ACTION,
                        payload={
                            NAME_KEY: message.payload[NAME_KEY],
                            (ERROR_KEY if isinstance(x, Exception) else RESULT_KEY): x,
                        },
                        event_id=message.event_id,
                    ),
                )

            self._is_initialized = True

    def register(self, name: Optional[str] = None, description: str = None, blocking: bool = False):
        """register function as action visiable on frontend page

        Args:
            name (Optional[str], optional): name of the action. Defaults to None(neetbox will use the function name when set to None).
            description (str, optional): description of the action. Defaults to None(neetbox will use function docs as default when set to None).
            blocking (bool, optional): whether to run the action in a blocked query. Defaults to False.

        Returns:
            Callable: the function itself.
        """
        if not self._is_initialized:
            self._initialize()
        return functools.partial(
            self._register, name=name, description=description, blocking=blocking
        )

    def _register(
        self, function: Callable, name: str = None, description: str = None, blocking: bool = False
    ):
        if (
            description is None and function.__doc__ is not None
        ):  # parse function doc as description
            description = function.__doc__
            if description:
                _description_lines = []
                for _line in description.split("\n"):
                    if len(_line):  # remove empty lines
                        _description_lines.append(_line)
                # find shortest lstrip
                min_lstrip = 99999
                for _line in _description_lines[1:]:  # skip first line
                    min_lstrip = min(len(_line) - len(_line.lstrip()), min_lstrip)
                _parsed_description = _description_lines[0] + "\n"
                for _line in _description_lines[1:]:
                    _parsed_description += _line[min_lstrip:] + "\n"
                description = _parsed_description

        packed = Action(function=function, name=name, description=description, blocking=blocking)
        self.ACTION_POOL._register(what=packed, name=packed.name, overwrite=True)
        connection.ws_send(
            event_type=EVENT_TYPE_NAME_STATUS,
            series="action",
            payload=self.get_action_dict(),
        )
        return function


actionManager = ActionManager()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240109

import os
import time
from threading import Thread

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.responses import FileResponse, RedirectResponse

import neetbox
from neetbox._protocol import *
from neetbox.logging import Logger, LogLevel

from .routers import project as project_router
from .routers import server as server_router

logger = Logger("FASTAPI", skip_writers_names=["ws"])
logger.log_level = LogLevel.DEBUG

serverapp = FastAPI()

front_end_dist_path = os.path.join(os.path.dirname(neetbox.__file__), "frontend_dist")
logger.info(f"using frontend dist path {front_end_dist_path}")

serverapp.include_router(
    project_router.crud_router,
    prefix=f"{API_ROOT}/{PROJECT_KEY}",
    tags=["project"],
)

serverapp.include_router(
    project_router.ws_router,
    prefix=f"{WS_ROOT}/{PROJECT_KEY}",
    tags=["websocket"],
)

serverapp.include_router(
    server_router.crud_router,
    prefix=f"{API_ROOT}/{SERVER_KEY}",
    tags=["server"],
)

serverapp.include_router(
    server_router.ws_router,
    prefix=f"{WS_ROOT}/{SERVER_KEY}",
    tags=["websocket"],
)

# mount web static files root
serverapp.mount("/static", StaticFiles(directory=front_end_dist_path), name="static")


@serverapp.get("/")
def redirect_to_web():
    return RedirectResponse(url="/web/", status_code=301)


@serverapp.get("/web/{path:path}")
async def serve_static_root(path: str):
    # Try to return the requested file
    path = "index.html" if not path else path
    static_file_path = f"{front_end_dist_path}/{path}"
    if os.path.isfile(static_file_path):
        return FileResponse(static_file_path)
    else:
        return FileResponse(f"{front_end_dist_path}/index.html")



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231206

from ._daemonable_process import DaemonableProcess
from ._messaging import messaging
from ._package import pipPackageHealper as pkg
from ._registry import Registry
from ._resource import ResourceLoader, download

__all__ = ["pkg", "ResourceLoader", "download", "DaemonableProcess", "Registry", "messaging"]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230416


def singleton(class_):
    class class_w(class_):
        _instance = None

        def __new__(cls, *args, **kwargs):
            if class_w._instance is None:
                class_w._instance = super(class_w, cls).__new__(cls, *args, **kwargs)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def patch(func):
    """Patch a function into a class type

    Args:
        func (Function): A funtion that takes at least one argument with a specific class type 'self:YourClass'

    Returns:
        funtion: patched function
    """
    cls = next(iter(func.__annotations__.values()))
    name = func.__defaults__[0]
    func.__qualname__ = f"{cls.__name__}.{func.__name__}"
    func.__module__ = cls.__module__
    if name is None:
        setattr(cls, func.__name__, func)
    else:
        func.__qualname__ = f"{cls.__name__}.{name}"
        setattr(cls, name, func)
    return func



from abc import ABC, abstractmethod

from neetbox.utils.mvc import Singleton


class ConfigInterface(ABC, metaclass=Singleton):
    @abstractmethod
    def __setattr__(self, key, value):
        ...

    @abstractmethod
    def __getattr__(self, key):
        ...

    @abstractmethod
    def update(self, another):
        ...

    @property
    @abstractmethod
    def here(self):
        ...



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231217

from time import time
from uuid import uuid4

from neetbox._protocol import *
from neetbox.utils.framing import get_caller_info_traceback
from neetbox.utils.massive import describe_object

from .._client import connection


class Progress:
    total: int
    done: int
    iterator: enumerate
    timestamp: str

    def __init__(self, input: Union[int, enumerate, any], *, name=None):
        """Decorate an iterable object, returning an iterator. Neetbox will send progress to frontend while you are iterating through it.

        Args:
            input (Union[int, enumerate, any]): Something to iterate or something to create enumeratable object.
        """
        self.name = name
        if isinstance(input, int):
            self.total = input
            self.iterator = iter(range(input))
        elif isinstance(input, enumerate):
            self.total = None
            self.iterator = input
        else:
            self.total = len(input)
            self.iterator = iter(input)
        self.caller_identity = get_caller_info_traceback(stack_offset=2)

        self.done = 0
        self.start_time = time()  # Track the start time
        self.timestamp = get_timestamp()

    def __enter__(self):
        # print(f"entering {self.caller_identity}")
        return self

    def __exit__(self, type, value, traceback):
        # print(f"leaving {self.caller_identity}")
        return

    def __iter__(self):
        return self

    def __next__(self):
        if self.total and self.done > self.total:
            raise StopIteration
        self.done += 1
        elapsed_time = time() - self.start_time
        rate = self.done / elapsed_time if elapsed_time > 0 else 0  # Calculate the iteration rate
        iter_next = next(self.iterator)

        connection.ws_send(
            event_type=EVENT_TYPE_NAME_PROGRESS,
            series=self.caller_identity.strid + (self.name or ""),
            payload={
                NAME_KEY: self.name or self.caller_identity.last_describable,
                "step": self.done,
                "current": describe_object(iter_next, length_limit=16),
                "total": self.total,
                "rate": rate if self.done > 1 else -1.0,
            },
            timestamp=self.timestamp,
            _history_len=1,
        )
        return iter_next

    def __len__(self):
        if self.total is None:
            raise RuntimeError("you are trying to get length of an enumerate object")
        return self.total


# if __name__ == "__main__":
#     from tqdm import tqdm
#     from time import sleep

#     with Progress(10) as p:
#         for i in p:
#             sleep(0.1)  # Simulate some work

#     with tqdm(Progress(range(10))) as p:
#         for i in p:
#             sleep(0.1)  # Simulate some work



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230319

import inspect
import types
from dataclasses import dataclass
from os import path
from typing import Union


def get_frame_traceback(stack_offset=1):
    stack = inspect.stack()
    stack_offset = len(stack) - 1 if stack_offset >= len(stack) else stack_offset
    return stack[stack_offset]


def get_frame_func_name_traceback(stack_offset=1):
    frame = get_frame_traceback(stack_offset + 1)
    func = frame.function
    return None if func == "<module>" else func


def get_frame_class_traceback(stack_offset=1):
    frame = get_frame_traceback(stack_offset + 1)
    _locals = frame[0].f_locals
    if "self" in _locals:
        return _locals["self"].__class__
    return None


def get_frame_module_traceback(stack_offset=1) -> Union[types.ModuleType, None]:
    frame = get_frame_traceback(stack_offset + 1)
    module = inspect.getmodule(frame[0])
    return module


def get_frame_filepath_traceback(stack_offset=1):
    frame = get_frame_traceback(stack_offset + 1)
    return frame.filename


@dataclass
class TracebackInfo:
    frame = None
    lineno: int = None
    func_name: str = None
    locals = None
    class_obj = None
    class_name: str = None
    module = None
    module_name: str = None
    filepath: str = None
    filename: str = None

    @classmethod
    def parse(cls, frame) -> "TracebackInfo":
        stacktrace = TracebackInfo()
        stacktrace.frame = frame
        stacktrace.lineno = frame.lineno
        stacktrace.func_name = frame.function if frame.function != "<module>" else None
        stacktrace.locals = frame[0].f_locals
        if "self" in stacktrace.locals:
            stacktrace.class_obj = stacktrace.locals["self"].__class__
            if stacktrace.class_obj:
                stacktrace.class_name = stacktrace.class_obj.__name__
        module: Union[types.ModuleType, None] = inspect.getmodule(frame[0])
        stacktrace.module = module
        stacktrace.module_name = module.__name__ if module else None
        stacktrace.filepath = path.abspath(frame.filename)
        stacktrace.filename = path.basename(frame.filename)
        return stacktrace

    @property
    def last_describable(self):
        for identity in [self.func_name, self.class_name, self.module_name, self.filepath, None]:
            if identity is not None:
                return identity

    @property
    def strid(self):
        return "-".join(self.as_str_sequence(lineno=True))

    @property
    def json(self):
        return {
            "file": f"{self.filepath}",
            "modlue": f"{self.module_name}",
            "class": f"{self.class_name}",
            "function": f"{self.func_name}",
        }

    def format(self, fmt: str):
        result = (
            fmt.replace(r"%F", self.filepath or "_")
            .replace(r"%m", self.module_name or "_")
            .replace(r"%c", self.class_name or "_")
            .replace(r"%f", self.func_name or "_")
            .replace(r"%l", str(self.lineno))
        )
        return result

    def as_str_sequence(self, lineno=False):
        return [
            prop
            for prop in [
                self.filepath,
                str(self.lineno) if lineno else None,
                self.module_name,
                self.class_name,
                self.func_name,
            ]
            if prop is not None
        ]

    def __eq__(self, __value: object) -> bool:
        assert isinstance(__value, TracebackInfo), f"cannot compare {__value} as a {TracebackInfo}"
        return (
            self.filepath == __value.filepath
            and self.module_name == __value.module_name
            and self.class_name == __value.class_name
            and self.func_name == __value.func_name
        )

    def __repr__(self) -> str:
        return "\n".join([f"{k}:\t{v}" for k, v in self.json.items()])


def get_caller_info_traceback(stack_offset=1):
    frame = get_frame_traceback(stack_offset + 1)
    return TracebackInfo.parse(frame)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230414


from neetbox._protocol import *

from ._client import connection


def _get(api, root=None):
    r = connection.get(api=api, root=root)
    _data = r.json()
    return _data


def _post(api, root=None, data=None):
    r = connection.post(api=api, root=root, json=data)
    return r


def shutdown(root=None):
    api = f"{API_ROOT}/{SERVER_KEY}/shutdown"
    return _post(api, root=root)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230413

import types
from typing import Union

from neetbox._protocol import MACHINE_ID_KEY, PROJECT_ID_KEY, RUN_ID_KEY

from .project import _get_module_level_config, export_default_config
from .user import get as get_user_config

__IS_WORKSPACE_LOADED = False


def get_module_level_config(module: Union[str, types.ModuleType] = None):
    global __IS_WORKSPACE_LOADED
    if not __IS_WORKSPACE_LOADED:
        __IS_WORKSPACE_LOADED = True
        from .project import _create_load_workspace

        _create_load_workspace()

    module_config = _get_module_level_config(module, stack_offset=3)
    return module_config


def get_machine_id():
    return get_user_config(MACHINE_ID_KEY)


def get_project_id():
    return get_module_level_config("@")[PROJECT_ID_KEY]


def get_run_id():
    return get_module_level_config("@")[RUN_ID_KEY]


__all__ = ["get_module_level_config", "export_default_config", "get_project_id", "get_run_id"]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240110

from typing import Dict
from uuid import uuid4

from fastapi import WebSocket
from rich import box
from rich.console import Console
from rich.table import Table

from neetbox._protocol import *
from neetbox.logging import Logger
from neetbox.server.fastapi.routers.project._bridge import Bridge
from neetbox.utils.mvc import Singleton

console = Console()
logger = Logger("Websocket.Project", skip_writers_names=["ws"])


@dataclass
class WSClient:
    id: str
    ws: WebSocket
    project_id: str
    identity_type: IdentityType
    run_id: str = None


class WSConnectionManager(metaclass=Singleton):
    id2client: Dict[str, WSClient]
    ws2client: Dict[WebSocket, WSClient]
    event_handlers: dict
    default_json_handler: dict

    def __init__(self) -> None:
        from ._event_type_handlers import (
            EVENT_TYPE_HANDLERS,
            on_event_type_default_json,
        )

        self.id2client = {}
        self.ws2client = {}
        self.event_handlers = EVENT_TYPE_HANDLERS
        self.default_json_handler = on_event_type_default_json
        logger.info(f"loaded event handlers: {list(EVENT_TYPE_HANDLERS.keys())}")

    async def handshake(self, websocket: WebSocket):
        await websocket.accept()
        id = str(uuid4())
        logger.info(f"websocket {id} connected, waiting for handshake...")
        message = await websocket.receive_text()  # the first message should be handshake message
        try:
            message = EventMsg.loads(message)
        except Exception as e:
            logger.err(f"error encountered '{e}' while handling handshake message: {message}")
            websocket.close(code=1003, reason="invalid handshake")
            return None
        assert (
            message.event_type == EVENT_TYPE_NAME_HANDSHAKE
        ), f"handshake expected but got {message.event_type}"
        assert message.project_id is not None, f"cannot handshake without project id"
        # assign this client to a Bridge
        logger.info(
            f"handling handshake for {message.identity_type}, project id '{message.project_id}'"
        )
        ws_client = WSClient(
            id=id, ws=websocket, project_id=message.project_id, identity_type=message.identity_type
        )
        if message.identity_type == IdentityType.WEB:
            if not Bridge.has(message.project_id):  # there is no such bridge
                merge_msg = {
                    PAYLOAD_KEY: {ERROR_KEY: 404, REASON_KEY: "project id not found"},
                    IDENTITY_TYPE_KEY: IdentityType.SERVER,
                }  # no such bridge, dropping...

            else:  # new connection from frontend
                bridge = Bridge.of_id(message.project_id)
                bridge.web_ws_list.append(ws_client)
                self.id2client[id] = ws_client
                self.ws2client[websocket] = ws_client
                merge_msg = {
                    PAYLOAD_KEY: {RESULT_KEY: 200, REASON_KEY: "join success"},
                    IDENTITY_TYPE_KEY: IdentityType.SERVER,
                }
        elif message.identity_type == IdentityType.CLI:
            # new connection from cli
            bridge = Bridge(project_id=message.project_id)
            run_id = message.run_id
            ws_client.run_id = run_id
            if run_id not in bridge.cli_ws_dict:
                bridge.cli_ws_dict[run_id] = ws_client  # assign cli to bridge
                self.id2client[id] = ws_client
                self.ws2client[websocket] = ws_client
                merge_msg = {
                    PAYLOAD_KEY: {RESULT_KEY: 200, REASON_KEY: "join success"},
                    IDENTITY_TYPE_KEY: IdentityType.SERVER,
                }  # handshake 200
            else:  # run id already exist
                merge_msg = {
                    PAYLOAD_KEY: {
                        ERROR_KEY: 400,
                        REASON_KEY: "client with run id already connected",
                    },
                    IDENTITY_TYPE_KEY: IdentityType.SERVER,
                }
        else:
            logger.warn(f"unknown type({message.identity_type}) of websocket trying to handshake")
            return  # unknown connection type, dropping...

        await websocket.send_json(data=EventMsg.merge(message, merge_msg).json)
        logger.ok(f"wsclient(id={id}) handshake succeed.")
        table = Table(title="Connected Websockets", box=box.MINIMAL_DOUBLE_HEAD, show_lines=True)
        table.add_column(PROJECT_ID_KEY, justify="center", style="magenta", no_wrap=True)
        table.add_column(IDENTITY_TYPE_KEY, justify="center", style="cyan")
        table.add_column(f"web:ws/cli:{RUN_ID_KEY}", justify="center", style="green")
        for _, _bridge in Bridge.items():
            table.add_row(_bridge.project_id, "", "")
            for cli_run_id, _ in _bridge.cli_ws_dict.items():
                table.add_row("", IdentityType.CLI, cli_run_id)
            for ws_web in _bridge.web_ws_list:
                table.add_row("", IdentityType.WEB, f"websocket id {ws_web.id}")
        console.print(table)
        return ws_client

    def disconnect(self, websocket: WebSocket):
        if websocket not in self.ws2client:
            return  # ignore if not handshaked
        ws_client = self.ws2client[websocket]
        id = ws_client.id
        project_id = ws_client.project_id
        identity_type = ws_client.identity_type
        bridge = Bridge.of_id(project_id)
        if not bridge:
            return  # do nothing if bridge has been deleted
        if identity_type == IdentityType.WEB:  # is web ws, remove from bridge's web ws list
            _new_web_ws_list = [c for c in Bridge.of_id(project_id).web_ws_list if c.id != id]
            bridge.web_ws_list = _new_web_ws_list
        elif (
            identity_type == IdentityType.CLI
        ):  # is cli ws, remove the run id from bridge's connected clients
            run_id = ws_client.run_id
            del bridge.cli_ws_dict[run_id]
        del self.id2client[id]
        del self.ws2client[websocket]
        logger.info(
            f"a {identity_type}(ws client id {id}) disconnected from project '{project_id}'"
        )

    async def handle_event_msg(self, websocket: WebSocket, message: EventMsg):
        ws_client = self.ws2client[websocket]
        if not message.identity_type:
            message.identity_type = ws_client.identity_type
        if message.identity_type != ws_client.identity_type:
            logger.warn(
                f"Illegal IdentityType: expect {ws_client.identity_type} but got {message.identity_type}"
            )
            return  # security check, identityType should match identityType

        # handle regular event types
        if message.event_type in self.event_handlers:
            for handler in self.event_handlers[message.event_type]:
                await handler(message)
        else:
            await self.default_json_handler(
                message=message,
                forward_to=IdentityType.OTHERS,
                save_history=True,
            )


manager = WSConnectionManager()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230417

import collections
import time
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from functools import partial
from threading import Thread
from typing import Any, Callable, Optional, Union
from uuid import uuid4

from neetbox.config import get_module_level_config
from neetbox.logging import logger
from neetbox.utils import Registry

__TIME_CTR_MAX_CYCLE = Decimal("99999.0")
__TIME_UNIT_SEC = Decimal("0.1")

_WATCH_QUERY_DICT = Registry("__pipeline_watch")
_LISTEN_QUERY_DICT = collections.defaultdict(lambda: {})
_UPDATE_VALUE_DICT = collections.defaultdict(lambda: {})

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"  # YYYY-MM-DDTHH:MM:SS.SSS

_DEFAULT_CHANNEL = str(
    uuid4()
)  # default watch and listen channel. users use this channel by default. default channel name varies on each start
SYSTEM_CHANNEL = "__system"  # values on this channel will upload via http client


@dataclass
class _WatchConfig(dict):
    name: str
    interval: Decimal
    initiative: bool
    channel: str = _DEFAULT_CHANNEL  # use channel to distinct those values to upload via http


class _WatchedFun:
    def __init__(self, func: Callable, cfg: _WatchConfig) -> None:
        self.func = func
        self.cfg = cfg

    def __call__(self, *args, **kwargs) -> Any:
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return f"{self.func.__name__}, {self.cfg}"


def __get(name: str, channel):
    _the_value = _UPDATE_VALUE_DICT[channel].get(name, default=None)
    if _the_value and "value" in _the_value:
        _the_value = _the_value["value"]
    return _the_value


def __update_and_get(name: str, *args, **kwargs):
    global _UPDATE_VALUE_DICT
    _watched_fun: _WatchedFun = _WATCH_QUERY_DICT[name]
    _watch_config = _watched_fun.cfg
    _channel = _watch_config.channel
    _the_value = _watched_fun(*args, **kwargs)
    _UPDATE_VALUE_DICT[_channel][name] = {
        "value": _the_value,
        "timestamp": datetime.now().strftime(DATETIME_FORMAT),
        "interval": None
        if _watch_config.interval is None
        else (_watch_config.interval * __TIME_UNIT_SEC),
    }

    def __call_listeners(_name: str, _value, _cfg: _WatchConfig):
        t0 = time.perf_counter()
        for _listener_name, _listener_func in _LISTEN_QUERY_DICT[_name].items():
            _listener_func(_value)
        t1 = time.perf_counter()
        delta_t = t1 - t0
        if _cfg.initiative:
            return
        _update_interval = _cfg.interval
        expected_time_limit = _update_interval * __TIME_UNIT_SEC
        if delta_t > expected_time_limit:
            logger.warn(
                f"Watched value {_name} takes longer time({delta_t:.8f}s) to update than it was expected    ({expected_time_limit}s)."
            )

    Thread(target=__call_listeners, args=(name, _the_value, _watch_config), daemon=True).start()

    return _the_value


def _watch(
    func: Callable,
    name: Optional[str],
    interval: float,
    initiative: bool = False,
    overwrite: bool = False,
    _channel: str = None,
):
    _channel = _channel or _DEFAULT_CHANNEL
    name = name or func.__name__
    _WATCH_QUERY_DICT._register(
        name=name,
        what=_WatchedFun(
            func=func,
            cfg=_WatchConfig(
                name,
                interval=None if interval is None else Decimal(str(interval)),
                initiative=initiative,
                channel=_channel,
            ),
        ),
        overwrite=overwrite,
        tags=_channel,
    )
    if initiative:  # initiatively update the value dict when the function was called manually
        logger.debug(
            f"added {name} to daemon monitor. It will update on each call of the function."
        )
        return partial(__update_and_get, name)
    else:
        _update_latency = Decimal(str(interval)) * __TIME_UNIT_SEC
        logger.debug(
            f"added {name} to daemon monitor. It will update every {_update_latency} second(s)."
        )
        return partial(__get, name, _channel)


def watch(
    name: str = None,
    interval: float = None,
    initiative: bool = True,
    overwrite: bool = False,
    _channel: str = None,
):
    # set interval to None if passively update
    interval = None if initiative else (interval or get_module_level_config()["updateInterval"])
    return partial(
        _watch,
        name=name,
        interval=interval,
        initiative=initiative,
        overwrite=overwrite,
        _channel=_channel,
    )


def _listen(
    func: Callable,  # the listener itself
    target: Union[str, Callable],  # user may pass a Callable or name to watch as target
    listener_name: Optional[str] = None,  # or user may pass a name to watch
    overwrite: bool = False,
):
    listener_name = listener_name or func.__name__
    if not isinstance(target, str):
        if type(target) is partial:  # solve target is a function with @watch
            if target.func in [__update_and_get, __get]:
                target = target.args[0]
        else:
            target = target.__name__

    if listener_name in _LISTEN_QUERY_DICT[target]:
        if not overwrite:
            logger.warn(
                f"There is already a listener called '{listener_name}' lisiting '{target}' If you want to overwrite, try to listen with 'overwrite=True'"
            )
            return func
        else:
            # logger.warn(
            #     f"There is already a listener called '{listener_name}' lisiting '{target}', overwriting."
            # )
            pass
    _LISTEN_QUERY_DICT[target][listener_name] = func
    logger.debug(f"{listener_name} is now lisiting to {target}.")
    return func


def listen(target: Union[str, Any], listener_name: Optional[str] = None, overwrite: bool = False):
    return partial(_listen, target=target, listener_name=listener_name, overwrite=overwrite)


def _update_thread():
    # update values
    _ctr = Decimal("0.0")
    while True:
        _ctr = (_ctr + __TIME_UNIT_SEC) % __TIME_CTR_MAX_CYCLE
        time.sleep(float(__TIME_UNIT_SEC))
        for _vname, _watched_fun in _WATCH_QUERY_DICT.items():
            _watch_config = _watched_fun.cfg
            if not _watch_config.initiative and _ctr % _watch_config.interval == Decimal(
                "0.0"
            ):  # do update
                _ = __update_and_get(_vname)


update_thread = Thread(target=_update_thread, daemon=True)
update_thread.start()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230413

import functools
import json
from typing import Any, Callable, Dict, List, Optional, Sequence, Union
from uuid import uuid4


class _RegEndpoint:
    def __init__(self, what, tags=None):
        """Generate a massive type which contains both the regietered object and it's tags

        Args:
            what (_type_): The object being registered
            tags (_type_, optional): The tags. Defaults to None.
        """
        self.what = what
        self.tags = tags

    def __str__(self) -> str:
        return f"{self.what} with tags: {self.tags}"


def _tags_match(search_tags, in_tags) -> bool:
    # check if all tags in f_tags are listed in s_tags
    if type(search_tags) is not list:
        search_tags = [search_tags]
    for _t in search_tags:
        if _t not in in_tags:
            return False
    return True


class Registry(dict):

    """Register Helper Class
    A Register is a 'dict[str:any]'
    Registers are stored in a pool of type dict[str:Register]
    """

    # class level
    _registry_pool: Dict[str, "Registry"] = dict()  # all registeres are stored here

    def __new__(cls, name: str) -> "Registry":
        if name in cls._registry_pool:
            return cls._registry_pool[name]
        instance = dict.__new__(cls)
        cls._registry_pool[name] = instance
        return instance

    def __init__(self, name) -> None:
        if "initialized" not in self:
            self["initialized"] = True
            self.name = name

    def _register(
        self,
        what: Any,
        name: Optional[str] = None,
        overwrite: Union[bool, Callable] = lambda x: x + f"_{uuid4()}",
        tags: Optional[Union[str, Sequence[str]]] = None,
    ):
        # if not (inspect.isfunction(what) or inspect.isclass(what)):
        #     logger.warn(f"Registering {type(what)}, which is not a class or a callable.")
        name = name or what.__name__
        if type(tags) is str:
            tags = [tags]
        _endp = _RegEndpoint(what, tags)
        if name in self.keys():
            if isinstance(overwrite, Callable):
                name = overwrite(name)
            elif overwrite == True:
                pass
            else:
                raise RuntimeError(f"Unknown overwrite type.")

            self[name] = _endp
        else:
            self[name] = _endp
        return what

    def register(
        self,
        *,
        name: Optional[str] = None,
        overwrite: Union[bool, Callable] = lambda x: x + f"_{uuid4()}",
        tags: Optional[Union[str, Sequence[str]]] = None,
    ):
        return functools.partial(self._register, name=name, overwrite=overwrite, tags=tags)

    @classmethod
    def find(
        cls,
        name: Optional[str] = None,
        tags: Optional[Union[str, List[str]]] = None,
    ):
        if not name and not tags:
            # logger.err(
            #     ValueError("Please provide at least the name or the tags you want to find."),
            #     reraise=True,
            # )
            pass
        results = []
        # filter name
        for reg_name, reg in cls._registry_pool.items():
            private_sign = "__"
            if not reg_name.startswith(private_sign):
                if not name:
                    results += [(_n, _o) for _n, _o in reg.items(_real_type=False)]
                elif name in reg:
                    results.append((name, reg[name]))

        # filter tags
        if type(tags) is not list:
            tags = [tags]

        results = {_name: _endp.what for _name, _endp in results if _tags_match(tags, _endp.tags)}
        return results

    def filter(self, tags: Optional[Union[str, Sequence[str]]] = None):
        results = {
            _name: _endp.what for _name, _endp in self._items() if _tags_match(tags, _endp.tags)
        }
        return results

    def __getitem__(self, __key: str) -> Any:
        _v = self.__dict__[__key]
        if type(_v) is _RegEndpoint:
            _v = _v.what
        return _v

    def get(self, key: str, **kwargs):
        if key in self.__dict__:
            _v = self.__dict__[key]
            if type(_v) is _RegEndpoint:
                _v = _v.what
            return _v
        else:
            if "default" in kwargs:
                return kwargs["default"]
            else:
                raise RuntimeError(f"key {key} not found")

    def __setitem__(self, k, v) -> None:
        self.__dict__[k] = v

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return [_item[0] for _item in self.__dict__.items() if type(_item[1]) is _RegEndpoint]

    def values(self):
        return [_item[1].what for _item in self.__dict__.items() if type(_item[1]) is _RegEndpoint]

    def items(self, _real_type=True):
        _legal_items = [_item for _item in self.__dict__.items() if type(_item[1]) is _RegEndpoint]
        if _real_type:
            _legal_items = [(_k, _v.what) for _k, _v in _legal_items if type(_v) is _RegEndpoint]
        return _legal_items

    def _items(self, _real_type=True):
        _legal_items = [_item for _item in self.__dict__.items() if type(_item[1]) is _RegEndpoint]
        if _real_type:
            _legal_items = [(_k, _v) for _k, _v in _legal_items if type(_v) is _RegEndpoint]
        return _legal_items

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def __cmp__(self, dict_):
        return self.__cmp__(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __str__(self) -> str:
        return json.dumps(
            dict(self.items()),
            indent=4,
            ensure_ascii=False,
            sort_keys=False,
            separators=(",", ":"),
            default=str,
        )

    __repr__ = __str__



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231201

from ._manager import manager
from .abc import FetchType, SortType

__all__ = ["FetchType", "SortType", "manager"]



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230413

import functools
import time
from threading import Thread
from typing import Callable, List

import GPUtil
import psutil
from GPUtil import GPU
from GPUtil import getAvailable as _getAvailableNvGPU

from neetbox.config import export_default_config, get_module_level_config
from neetbox.extension import on_workspace_loaded
from neetbox.utils.mvc import Singleton


class CpuStatus:
    def __init__(self, id=-1, percentage=0.0, frequency=0.0):
        self.id = id
        self.percentage = percentage
        self.frequency = frequency

    def __str__(self) -> str:
        return f"CPU({self.id}), {(self.percentage)}%, {(self.frequency)}Mhz"

    @property
    def json(self):
        return {"id": self.id, "percentage": self.percentage, "frequency": self.frequency}


class CpuStatistics:
    def __init__(self, ctx_switches, interrupts, soft_interrupts, syscalls) -> None:
        self.ctx_switches = ctx_switches
        self.interrupts = interrupts
        self.soft_interrupts = soft_interrupts
        self.syscalls = syscalls

    @property
    def json(self):
        return {
            "ctxSwitches": self.ctx_switches,
            "interrupts": self.interrupts,
            "softInterrupts": self.soft_interrupts,
            "syscalls": self.syscalls,
        }

    def __str__(self) -> str:
        result = ""
        for k, v in self.json.items():
            result += f"{k}:\t{v}"
        return result


class MemoryStatus:
    def __init__(self, total, available, used, free) -> None:
        self.total = total
        self.available = available
        self.used = used
        self.free = free

    @property
    def json(self):
        return {
            "total": self.total,
            "available": self.available,
            "used": self.used,
            "free": self.free,
        }

    def __str__(self) -> str:
        result = ""
        for k, v in self.json.items():
            result += f"{k}:\t{v}"
        return result


class NvGpuStatus(GPU):
    @classmethod
    def parse(cls, other: GPU):
        return NvGpuStatus(
            ID=other.id,
            uuid=other.uuid,
            load=other.load,
            memoryTotal=other.memoryTotal,
            memoryUsed=other.memoryUsed,
            memoryFree=other.memoryFree,
            driver=other.driver,
            gpu_name=other.name,
            serial=other.serial,
            display_mode=other.display_mode,
            display_active=other.display_active,
            temp_gpu=other.temperature,
        )

    @property
    def json(self):
        return {
            "id": self.id,
            "name": self.name,
            "load": self.load,
            "memoryUtil": self.memoryUtil,
            "memoryTotal": self.memoryTotal,
            "memoryFree": self.memoryFree,
            "memoryUsed": self.memoryUsed,
            "temperature": self.temperature,
            "driver": self.driver,
        }

    def __str__(self) -> str:
        return super().__str__()


class Hardware(metaclass=Singleton):
    watch_thread: Thread = None
    _do_watch: bool = True
    _update_interval: float = 1.0

    # properties
    _cpus: List[CpuStatus]
    _gpus: List[NvGpuStatus]
    _cpu_statistics: CpuStatistics
    _memory: MemoryStatus

    def __init__(self) -> None:
        self._cpus = [CpuStatus() for _ in range(psutil.cpu_count(logical=True))]
        self._cpu_statistics = CpuStatistics(*psutil.cpu_stats())
        self._gpus = [NvGpuStatus.parse(gpu) for gpu in GPUtil.getGPUs()]
        self._with_gpu = False if len(self._gpus) == 0 else True
        virtual_memory = psutil.virtual_memory()
        self._memory = MemoryStatus(
            total=virtual_memory[0] / 1e6,
            available=virtual_memory[1] / 1e6,
            used=virtual_memory[3] / 1e6,
            free=virtual_memory[4] / 1e6,
        )

    @property
    def cpus(self):
        return self._cpus

    @property
    def cpu_statistics(self):
        return self._cpu_statistics

    @property
    def memory(self):
        return self._memory

    @property
    def gpus(self):
        return self._gpus

    @property
    def with_gpu(self):
        return self._with_gpu

    @property
    def json(self):
        return {
            "cpus": [cpu.json for cpu in self._cpus],
            "cpustat": self._cpu_statistics.json,
            "ram": self._memory.json,
            "gpus": [gpu.json for gpu in self._gpus],
        }

    _on_update_call_backs = []

    def add_on_update_call_back(self, func: Callable):
        self._on_update_call_backs.append(func)

    def set_start_update_intervel(self, intervel=1.0) -> None:
        if intervel < 0:
            self._do_watch = False
            return
        self._do_watch = True
        self._update_interval = intervel
        if not self.watch_thread or not self.watch_thread.is_alive():

            def watch_hardware(do_update_gpus: bool):
                while self._do_watch:
                    # update cpu usage
                    cpu_percent = psutil.cpu_percent(percpu=True)
                    cpu_freq = psutil.cpu_freq(percpu=True)
                    if len(cpu_freq) == 1:
                        cpu_freq = cpu_freq * len(cpu_percent)
                    for index in range(len(cpu_percent)):
                        self._cpus[index] = CpuStatus(
                            id=index,
                            percentage=cpu_percent[index],
                            frequency=cpu_freq[index],
                        )
                    # update memory usage
                    virtual_memory = psutil.virtual_memory()
                    self._memory = MemoryStatus(
                        total=virtual_memory[0] / 1e6,
                        available=virtual_memory[1] / 1e6,
                        used=virtual_memory[3] / 1e6,
                        free=virtual_memory[4] / 1e6,
                    )
                    # update gpu usage
                    if do_update_gpus:
                        self._gpus = [NvGpuStatus.parse(gpu) for gpu in GPUtil.getGPUs()]
                    self._cpu_statistics = CpuStatistics(*psutil.cpu_stats())
                    for callback in self._on_update_call_backs:
                        callback(self.json)
                    time.sleep(self._update_interval)

            self.watch_thread = Thread(target=watch_hardware, args=(self._with_gpu,), daemon=True)
            self.watch_thread.start()


Hardware.get_available_NvGPU = _getAvailableNvGPU

# singleton
hardware = Hardware()


@export_default_config
def return_default_config() -> dict:
    return {"interval": 2}


# watch updates in daemon
@on_workspace_loaded(name="hardware-monit")
def load_monit_hardware():
    cfg = get_module_level_config()
    if cfg["interval"] > 0:  # if do monit hardware
        from neetbox._protocol import EVENT_TYPE_NAME_HARDWARE
        from neetbox.client import connection

        def ws_send_on_update(stat_json):
            connection.ws_send(
                event_type=EVENT_TYPE_NAME_HARDWARE,
                payload=stat_json,
                _history_len=1000,
            )

        hardware.add_on_update_call_back(ws_send_on_update)
        hardware.set_start_update_intervel(cfg["interval"])



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230318

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from neetbox._protocol import *
from neetbox.utils.framing import TracebackInfo


@dataclass
class LogStyle:
    datetime_format: Optional[str] = r"%Y-%m-%dT%H:%M:%S.%f"
    caller_info_format: Optional[str] = r"%m/%c/%f"


@dataclass
class RawLog:
    message: str
    caller_info: TracebackInfo
    caller_name_alias: Optional[str] = None
    timestamp: datetime = datetime.now()
    series: Optional[str] = None
    style: LogStyle = field(default_factory=LogStyle)  # fix python 3.11 dataclass issue

    @property
    def timestamp_formatted(self):
        return (
            self.timestamp.strftime(self.style.datetime_format)
            if self.style.datetime_format
            else ""
        )

    @property
    def caller_info_formatted(self):
        if self.caller_name_alias:
            return str(self.caller_name_alias)
        return (
            self.caller_info.format(self.style.caller_info_format)
            if self.style.caller_info_format
            else ""
        )

    @property
    def json(self) -> dict:
        return {
            SERIES_KEY: self.series,
            MESSAGE_KEY: self.message,
            TIMESTAMP_KEY: self.timestamp_formatted,
            CALLER_ID_KEY: self.caller_info_formatted,
        }

    def __repr__(self) -> str:
        result = ""
        for k, v in self.json.items():
            result += f"{k}: {v}\n"
        return result



import torch
import torch.nn.functional as F


class MaskEdgeDetecter(torch.nn.Module):
    """generate the 'edge bar' for a 0-1 mask Groundtruth of a image
    Algorithm is based on 'Morphological Dilation and Difference Reduction'

    Which implemented with fixed-weight Convolution layer with weight matrix looks like a cross,
    for example, if kernel size is 3, the weight matrix is:
        [[0, 1, 0],
        [1, 1, 1],
        [0, 1, 0]]

    """

    def __init__(self, kernel_size=3) -> None:
        super().__init__()
        self.kernel_size = kernel_size

    def _dilate(self, image, kernel_size=3):
        """Doings dilation on the image

        Args:
            image (_type_): 0-1 tensor in shape (B, C, H, W)
        """
        assert kernel_size % 2 == 1, "Kernel size must be odd"
        assert (
            image.shape[2] > kernel_size and image.shape[3] > kernel_size
        ), "Image must be larger than kernel size"

        kernel = torch.zeros((1, 1, kernel_size, kernel_size))
        kernel[0, 0, kernel_size // 2 : kernel_size // 2 + 1, :] = 1
        kernel[0, 0, :, kernel_size // 2 : kernel_size // 2 + 1] = 1
        kernel = kernel.float()
        res = F.conv2d(
            image,
            kernel.view([1, 1, kernel_size, kernel_size]),
            stride=1,
            padding=kernel_size // 2,
        )
        return (res > 0) * 1.0

    def _find_edge(self, image, kernel_size=3, return_all=False):
        """Find 0-1 edges of the image

        Args:
            image (_type_): 0-1 ndarray in shape (B, C, H, W)
        """
        image = image.clone().float()
        shape = image.shape

        if len(shape) == 2:
            image = image.reshape([1, 1, shape[0], shape[1]])
        if len(shape) == 3:
            image = image.reshape([1, shape[0], shape[1], shape[2]])
        assert image.shape[1] == 1, "Image must be single channel"

        img = self._dilate(image, kernel_size=kernel_size)

        erosion = self._dilate(1 - image, kernel_size=kernel_size)

        diff = -torch.abs(erosion - img) + 1
        diff = (diff > 0) * 1.0
        # res = dilate(diff)
        diff = diff.numpy()
        if return_all:
            return diff, img, erosion
        else:
            return diff

    def forward(self, x, return_all=False):
        """
        Args:
            image (_type_): 0-1 ndarray in shape (B, C, H, W)
        """
        return self._find_edge(x, self.kernel_size, return_all=return_all)



import argparse
import json
import sys


def run(argv):
    if len(argv) <= 1:
        print("_daemon_: Warning: empty daemon_config")
        daemon_config = None
    else:
        ap = argparse.ArgumentParser()
        ap.add_argument("--config")
        args = ap.parse_args()
        print(type(args.config))
        print(args.config)
        daemon_config = json.loads(args.config)
        print("Daemon started with config:", daemon_config)
    from .server_process import server_process

    server_process(daemon_config)


if __name__ == "__main__":
    print("daemon server starting.\n name:\t", __name__, "\nargs:\t", sys.argv)
    run(sys.argv)
    print("daemon server exiting.")



import importlib
from typing import Union

import pip

from neetbox.utils.framing import get_caller_info_traceback
from neetbox.utils.mvc import Singleton


class PipPackageHealper(metaclass=Singleton):
    def __init__(self) -> None:
        self.installed_packages = None

    def install(self, package, terminate=False):
        caller = get_caller_info_traceback(3)
        caller_name = caller.module_name if caller.module else caller.filename
        retry = 4
        _installed = False
        while retry:
            if not retry:
                error_str = "Bad Input"
                raise ValueError(error_str)
            print(f"{caller_name} want to install {package} via pip.")
            choice = input("Make your choice: [y]/n")
            choice = choice or "y"
            if choice in ["y", "yes"]:  # user choose to install
                print(f"installing {package} via pip...")
                pip.main(["install", package])
                _installed = True
                break
            if choice in ["n", "no"]:  # user choose not to install
                if terminate:  # the package must be installed
                    error_str = (
                        f"{caller_name} requires '{package}' but it is not going to be installed."
                    )
                    raise RuntimeError(error_str)
                else:
                    print(f"{package} is not going to be installed")
                    break
            else:  # illegal input
                retry -= 1
                if retry:
                    print(
                        f"illegal input: required 'y'/'n' but recieved '{choice}'. {retry} retries remaining."
                    )
        return _installed

    def is_installed(self, package: str, try_install_if_not: Union[str, bool] = True):
        caller = get_caller_info_traceback(3)
        caller_name = caller.module_name if caller.module else caller.filename
        if not self.installed_packages:
            self.installed_packages = []
            package = str(package)
        if package in self.installed_packages:
            return True
        try:
            importlib.import_module(package)
            self.installed_packages.append(package)
            return True
        except:  # noqa
            package_name_install = (
                package if type(try_install_if_not) is bool else try_install_if_not
            )
            print(f"{caller_name} requires '{package_name_install}' which is not installed.")
            if try_install_if_not:
                return self.install(package=package_name_install, terminate=True)
            return False


# singleton
pipPackageHealper = PipPackageHealper()



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20230315

import functools
import os
from datetime import date
from enum import Enum
from typing import Callable, Optional, Union

from neetbox.utils import Registry
from neetbox.utils.framing import get_caller_info_traceback

from ._formatting import LogStyle, RawLog
from .writers import FileLogWriter

LogWriters = Registry("LOG_WRITERS")


class LogLevel(Enum):
    ALL = 4
    DEBUG = 3
    INFO = 2
    WARNING = 1
    ERROR = 0

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value


class Logger:
    # global static
    _IDENTITY2LOGGER = {}

    def __init__(
        self,
        name_alias: str = None,
        style: Optional[LogStyle] = LogStyle(),
        log_level: LogLevel = LogLevel.INFO,
        skip_writers_names=[],
    ):
        """create a new logger

        Args:
            name_alias (str, optional): the name alias of the logger. Defaults to None.
            style (Optional[LogStyle], optional): logger's default log style. Defaults to LogStyle().
            log_level (LogLevel, optional): log level. Defaults to LogLevel.INFO.
            skip_writers_names (list, optional): names of writers to be skipped. Defaults to [].
        """

        self.name_alias = name_alias
        self._default_style = style
        self._log_level = log_level
        self.private_writers = {}
        self.skipped_writers_names = skip_writers_names

    def __new__(cls, name_alias: str = None, *args, **kwargs) -> "Logger":
        if name_alias in Logger._IDENTITY2LOGGER:
            return Logger._IDENTITY2LOGGER[name_alias]
        instance = super().__new__(cls)
        Logger._IDENTITY2LOGGER[name_alias] = instance
        return instance

    @property
    def style(self):
        return self._default_style

    @style.setter
    def style(self, style: LogStyle):
        self._default_style = style or LogStyle()

    @property
    def log_level(self):
        return self._log_level

    @log_level.setter
    def log_level(self, level: Union[LogLevel, str]):
        if type(level) is str:
            level = {
                "ALL": LogLevel.ALL,
                "DEBUG": LogLevel.DEBUG,
                "INFO": LogLevel.INFO,
                "WARNING": LogLevel.WARNING,
                "ERROR": LogLevel.ERROR,
            }[level]
        if type(level) is int:
            assert level >= 0 and level <= 3
            level = LogLevel(level)
        self._log_level = level

    def writer(self, name: str):
        def _add_private_writer(name, writer_func: Callable):
            if name in self.private_writers:
                self.warn(f"writer {name} already exist as a private writer. overwrite anyway.")
            self.private_writers[name] = writer_func
            return writer_func

        return functools.partial(_add_private_writer, name=name)

    def skip_writer_name(self, name: str):
        self.skipped_writers_names.append(name)

    def log(
        self,
        *content,
        series: Optional[str] = None,
        skip_writers_names: list[str] = [],
        stack_offset=2,
    ):
        # converting passed message args into a single string
        message = ""
        for msg in content:
            message += str(msg) + " "

        log = RawLog(
            message=message,
            caller_info=get_caller_info_traceback(stack_offset=stack_offset),
            caller_name_alias=self.name_alias,
            series=series,
            style=self._default_style,
        )

        writers = []
        for writer_name, writer_func in LogWriters.items():  # collect global writers
            if writer_name not in self.skipped_writers_names + skip_writers_names:
                writers.append((writer_name, writer_func))

        for writer_name, writer_func in self.private_writers.items():  # collect private writers
            if writer_name not in self.skipped_writers_names + skip_writers_names:
                writers.append((writer_name, writer_func))

        for writer_name, writer_func in writers:
            try:
                writer_func(log)
            except Exception as e:
                print(f"log writer {writer_name} fialed: {e}, original message:")
                print(log)

        return self

    def ok(
        self,
        *content,
        skip_writers_names: list[str] = [],
        stack_offset=2,
    ):
        if self._log_level >= LogLevel.INFO:
            self.log(
                *content,
                series=f"ok",
                skip_writers_names=skip_writers_names,
                stack_offset=stack_offset + 1,
            )
        return self

    def debug(
        self,
        *content,
        skip_writers_names: list[str] = [],
        stack_offset=2,
    ):
        if self._log_level >= LogLevel.INFO:
            self.log(
                *content,
                series=f"debug",
                skip_writers_names=skip_writers_names,
                stack_offset=stack_offset + 1,
            )
        return self

    def info(
        self,
        *content,
        skip_writers_names: list[str] = [],
        stack_offset=2,
    ):
        if self._log_level >= LogLevel.INFO:
            self.log(
                *content,
                series=f"info",
                skip_writers_names=skip_writers_names,
                stack_offset=stack_offset + 1,
            )
        return self

    def warn(
        self,
        *content,
        skip_writers_names: list[str] = [],
        stack_offset=2,
    ):
        if self._log_level >= LogLevel.INFO:
            self.log(
                *content,
                series=f"warning",
                skip_writers_names=skip_writers_names,
                stack_offset=stack_offset + 1,
            )
        return self

    def err(
        self,
        err,
        skip_writers_names: list[str] = [],
        stack_offset=2,
        reraise=False,
    ):
        if self._log_level >= LogLevel.ERROR:
            self.log(
                str(err),
                series=f"error",
                skip_writers_names=skip_writers_names,
                stack_offset=stack_offset + 1,
            )
        if reraise:
            if not isinstance(err, Exception):
                err = RuntimeError(err)
            raise err
        return self

    def mention(
        self,
        mention_args=True,
        mention_result=True,
        skip_writers_names=[],
    ):
        def with_logging(func):
            @functools.wraps(func)
            def _with_logging(*args, **kwargs):
                self.log(
                    f"Entering: {func.__name__}" + f", args={args}, kwargs={kwargs}"
                    if mention_args
                    else "",
                    series=f"mention",
                    skip_writers_names=skip_writers_names,
                    stack_offset=4,
                )
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    raise e
                self.log(
                    f"Leaving: {func.__name__}" + f", with returned value {result}"
                    if mention_result
                    else "",
                    series=f"mention",
                    skip_writers_names=skip_writers_names,
                    stack_offset=4,
                )
                return result

            return _with_logging

        return with_logging

    def set_log_dir(self, path, dedicated_file=False):
        if os.path.isfile(path):
            raise FileExistsError("Target path is not a directory.")
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except Exception:
                raise OSError(f"Failed when trying to create directory {path}")
        filename = ""
        filename += self.name_alias or "" if dedicated_file else ""
        filename += str(date.today()) + ".log"
        file_writer = FileLogWriter(os.path.join(path, filename))
        self.private_writers["file"] = file_writer.write
        return self


DEFAULT_LOGGER = Logger(None)



# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20240110

from collections import defaultdict
from typing import Callable

from neetbox._protocol import *

from .._bridge import Bridge

EVENT_TYPE_HANDLERS = defaultdict(list)


def on_event(type_name: str, *args):
    def _on_event(func: Callable) -> Callable:
        EVENT_TYPE_HANDLERS[type_name].append(func)
        return func

    return _on_event


async def on_event_type_default_json(
    message: EventMsg,
    forward_to: IdentityType = IdentityType.OTHERS,
    save_history=True,
):
    bridge = Bridge.of_id(message.project_id)
    if save_history:
        message.id = bridge.save_json_to_history(
            table_name=message.event_type,
            json_data=message.payload,
            series=message.series,
            run_id=message.run_id,
            timestamp=message.timestamp,
            num_row_limit=message.history_len,
        )
    if forward_to:
        if forward_to == IdentityType.SELF:
            forward_to = message.identity_type
        if forward_to == IdentityType.OTHERS:
            forward_to = (
                IdentityType.WEB if message.identity_type == IdentityType.CLI else IdentityType.CLI
            )
        if forward_to in [IdentityType.WEB, IdentityType.BOTH]:
            await bridge.ws_send_to_frontends(message)  # forward to frontends
        elif forward_to in [IdentityType.CLI, IdentityType.BOTH]:
            await bridge.ws_send_to_client(message, run_id=message.run_id)  # forward to frontends

    return  # return after handling log forwardin


@on_event(EVENT_TYPE_NAME_STATUS)
async def on_event_type_status(message: EventMsg):
    bridge = Bridge.of_id(message.project_id)
    bridge.set_status(run_id=message.run_id, series=message.series, value=message.payload)


@on_event(EVENT_TYPE_NAME_HPARAMS)
async def on_event_type_hyperparams(message: EventMsg):
    bridge = Bridge.of_id(message.project_id)
    current_hyperparams = bridge.get_status(
        run_id=message.run_id, series=EVENT_TYPE_NAME_HPARAMS
    )  # get hyper params from status
    if message.series:  # if series of hyperparams specified
        current_hyperparams[message.series] = message.payload
    else:
        for k, v in message.payload.items():
            current_hyperparams[k] = v
    bridge.set_status(
        run_id=message.run_id, series=EVENT_TYPE_NAME_HPARAMS, value=current_hyperparams
    )


@on_event(EVENT_TYPE_NAME_ACTION)
async def on_event_type_action(message: EventMsg):
    await on_event_type_default_json(
        message=message, forward_to=IdentityType.OTHERS, save_history=False
    )
